/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * netio.{cc,hh} -- Read packets from OSKit netio, and pass along.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "fromnetio.hh"
#include "error.hh"
#include "packet.hh"
#include "confparse.hh"
#include "glue.hh"
#include <unistd.h>
#include "../standard/scheduleinfo.hh"
#include "elements/oskit/netio_guts.h"
#ifdef  MULTITHREADED
#include "router.hh"
#endif

#ifdef MEASURE
#include <oskit/machine/proc_reg.h>
#ifdef PERFMON
#include <oskit/machine/perfmon.h>
#endif
#endif

inline void
FromNetIO::enq(Packet *p)
{
	assert(p);
	
	int next = next_i(_qtail);
	if (next != _qhead) {
		_q[_qtail] = p;
		_qtail = next;
	} else {
		p->kill();
		_qdrops++;
	}
}

inline Packet *
FromNetIO::deq()
{
	if (_qhead != _qtail) {
		Packet *p = _q[_qhead];
		assert(p);
		_qhead = next_i(_qhead);
		return p;
	} else
		return 0;
}

FromNetIO::FromNetIO()
  : _ifnum(-1), _recvs(0), _qhead(0), _qtail(0), _max(512), _qdrops(0)
{
	add_output();
}

FromNetIO::~FromNetIO()
{
	if (_ifnum >= 0) {
		(void) oskit_destroy_fromnetio(_ifnum, _priv);

		if (_recvs)
			click_chatter("FromNetIO::~FromNetIO(%d): "
				      "%d packets recevied, %d dropped due to "
				      "Q overflow", _ifnum, _recvs, _qdrops);
	}
}

FromNetIO *
FromNetIO::clone() const
{
	return new FromNetIO;
}

int
FromNetIO::configure(const String &conf, ErrorHandler *errh)
{
	if (cp_va_parse(conf, this, errh,
			cpInteger, "interface number", &_ifnum,
			cpEnd) < 0)
		return -1;

	return 0;
}

int
FromNetIO::initialize(ErrorHandler *errh)
{
	int	err;

	err = oskit_create_fromnetio(this, 0, _ifnum, &_priv);
	if (err != 0)
		return err;

	_q = new Packet *[_max + 1];
	if (_q == 0)
		return errh->error("FromNetIO::initialize: out of memory");

	_qhead = _qtail = 0;

#ifndef  INTERRUPT_DRIVEN
	/*
	 * When not INTERRUPT_DRIVEN, the element is always scheduled to
	 * run and look for incoming packets.
	 */
	ScheduleInfo::join_scheduler(this, errh);
#endif
	return 0;
}

int
FromNetIO::got_packet(struct sk_buff *skb)
{
#ifdef MEASURE
	unsigned long long now = get_tsc();
#endif
	_recvs++;
	
	/*
	 * Make a packet out of the bufio data. If this succeeds, the
	 * packet constructor will bump the bufio reference count.
	 */
	Packet* p = Packet::make(skb);

#ifdef MEASURE
#ifdef PERFMON
	p->set_measure_counter0(read_perfcounter(0));
	p->set_measure_counter1(read_perfcounter(1));
#endif
/*	unsigned long long now = get_tsc(); */
	p->set_measure_measure();
	p->set_measure_created(now);
#endif
	/*
	 * Add to the queue. The packet will be dropped if the queue fills.
	 */
	enq(p);
#ifdef  INTERRUPT_DRIVEN
	if (!scheduled())
		reschedule();
#endif
#ifdef  MULTITHREADED
	this->router()->wakeup();
#endif
	
	return 0;
}

void
FromNetIO::run_scheduled()
{
	Packet *p;
	int	i;

#ifdef MEASURE
#define MAXGROUP	1
#else
#define MAXGROUP	8
#endif

	/*
	 * Fire off in small groups of packets to avoid click driver
	 * loop overhead.
	 */
	for (i = 0; i < MAXGROUP; i++) {
		p = deq();
		if (!p)
			break;
#ifdef MEASURE0
#ifdef PERFMON
		p->set_measure_counter0(read_perfcounter(0));
		p->set_measure_counter1(read_perfcounter(1));
#endif
		p->set_measure_measure();
		p->set_measure_created(get_tsc());
#endif
		output(0).push(p);
	}

#ifdef  INTERRUPT_DRIVEN
	if (! qempty())
#endif
		reschedule();
}

/*
 * This is the upcall from the netio code.
 */
extern "C" int
click_FromNetio_in(void *obj, struct sk_buff *skb, int)
{
	FromNetIO	*netio = (FromNetIO *) obj;

	return netio->got_packet(skb);
}

EXPORT_ELEMENT(FromNetIO)
