/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * netio.{cc,hh} -- Poll packets from OSKit netio, and pass along.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "pollnetio.hh"
#include "error.hh"
#include "packet.hh"
#include "confparse.hh"
#include "glue.hh"
#include <unistd.h>
#include "../standard/scheduleinfo.hh"
#include "elements/oskit/netio_guts.h"
#ifdef  MULTITHREADED
#include "router.hh"
#endif

#ifdef MEASURE
#include <oskit/machine/proc_reg.h>
#ifdef PERFMON
#include <oskit/machine/perfmon.h>
#endif
#endif

PollNetIO::PollNetIO()
  : _ifnum(-1)
{
	add_output();
}

PollNetIO::~PollNetIO()
{
	if (_ifnum >= 0) {
		(void) oskit_destroy_fromnetio(_ifnum, _priv);
	}
}

PollNetIO *
PollNetIO::clone() const
{
	return new PollNetIO;
}

int
PollNetIO::configure(const String &conf, ErrorHandler *errh)
{
	if (cp_va_parse(conf, this, errh,
			cpInteger, "interface number", &_ifnum,
			cpEnd) < 0)
		return -1;

	return 0;
}

int
PollNetIO::initialize(ErrorHandler *errh)
{
	int	err;

	err = oskit_create_fromnetio(this, 1, _ifnum, &_priv);
	if (err != 0)
		return err;

	ScheduleInfo::join_scheduler(this, errh);

	return 0;
}

void
PollNetIO::run_scheduled()
{
#define MAXGROUP	12
	Packet			*p;
	int			i;
	struct sk_buff		*skbs[MAXGROUP], *skb;
	int			count = MAXGROUP;
#ifdef MEASURE
	unsigned long long	now = get_tsc();
#endif

	/*
	 * Fire off in small groups of packets to avoid click driver
	 * loop overhead.
	 */
	oskit_netio_rx_poll(_priv, &count, skbs);

	for (i = 0; i < count; i++) {
		skb = skbs[i];
		
#ifdef MEASURE
		now = get_tsc();
#endif
		p = Packet::make(skb);
		assert(p);
#ifdef MEASURE
#ifdef PERFMON
		p->set_measure_counter0(read_perfcounter(0));
		p->set_measure_counter1(read_perfcounter(1));
#endif
		p->set_measure_measure();
		p->set_measure_created(now);
#endif
		output(0).push(p);
	}

	reschedule();
}

EXPORT_ELEMENT(PollNetIO)
