/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * tonetio.{cc,hh} -- Send packets to a OSKit netio.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "tonetio.hh"
#include "error.hh"
#include "packet.hh"
#include "confparse.hh"
#include "glue.hh"
#include <unistd.h>
#include "elements/oskit/netio_guts.h"
#include "elements/standard/scheduleinfo.hh"
#include <oskit/machine/perfmon.h>
#include <oskit/machine/proc_reg.h>

#ifdef	MEASURE
int			packet_count;
unsigned long long	packet_total;
int			packet_flag;
#ifdef  PERFMON
unsigned long long	perfmon_counter0;
unsigned long long	perfmon_counter1;
#endif
#endif

ToNetIO::ToNetIO()
  : Element(1, 0), _ifnum(-1), _drops(0), _sent(0)
{
}

ToNetIO::~ToNetIO()
{
	if (_ifnum >= 0) {
		(void) oskit_destroy_tonetio(_ifnum, _priv);

		if (_sent)
			click_chatter("ToNetIO::~ToNetIO(%d): "
				      "%d packets sent, %d dropped due to "
				      "Q overflow", _ifnum, _sent, _drops);
	}
}

ToNetIO *
ToNetIO::clone() const
{
	return new ToNetIO;
}

int
ToNetIO::configure(const String &conf, ErrorHandler *errh)
{
	if (cp_va_parse(conf, this, errh,
			cpInteger, "interface number", &_ifnum,
			cpEnd) < 0)
		return -1;

	return 0;
}

int
ToNetIO::initialize(ErrorHandler *errh)
{
	int	foo;

	if ((foo = oskit_create_tonetio(this, _ifnum, &_priv)))
		return foo;
	
	if (input_is_pull(0))
		ScheduleInfo::join_scheduler(this, errh);
	
	return 0;
}

void
ToNetIO::uninitialize(void)
{
	unschedule();
}

void
ToNetIO::push(int, Packet *p)
{
#ifdef  MEASURE0
	unsigned long long now = get_tsc();
#ifdef  PERFMON
	unsigned long long counter0 = read_perfcounter(0);
	unsigned long long counter1 = read_perfcounter(1);
#endif
#endif

#ifdef  MEASURE
	unsigned long long now = get_tsc();
#endif
	_sent++;
	if (oskit_sendto_netio_withbufio(_priv, p->skb(), p->length()))
		_drops++;
	p->kill();

#ifdef	MEASURE
	/*
	 * Record time through the router.
	 */
	if (p->get_measure_measure()) {
		if (packet_count == 100 && packet_flag == 0) {
			packet_flag  = 1;
			packet_count = 0;
			packet_total = 0;
#ifdef  PERFMON
			perfmon_counter0 = 0;
			perfmon_counter1 = 0;
#endif
		}
		packet_count++;
		packet_total += (now - p->get_measure_created());
#ifdef  PERFMON
		perfmon_counter0 += (counter0 - p->get_measure_counter0());
		perfmon_counter1 += (counter1 - p->get_measure_counter1());
#endif
	}
#endif
}

bool
ToNetIO::wants_packet_upstream() const
{
	return input_is_pull(0);
}

void
ToNetIO::run_scheduled()
{
	while (Packet *p = input(0).pull())
		push(0, p);
	reschedule();
}

EXPORT_ELEMENT(ToNetIO)
