/*
 * packet.{cc,hh} -- a packet structure. In the Linux kernel, a synonym for
 * `struct sk_buff'
 * Eddie Kohler, Robert Morris
 *
 * Copyright (c) 1999-2000 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYRIGHT' file in the source
 * distribution.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "packet.hh"
#include "glue.hh"
#include <stdarg.h>
#include <unistd.h>
#include <assert.h>

#ifdef CLICK_OSKIT
extern "C" {
void panic(const char *__format, ...);
}
#endif

#if defined(__KERNEL__) || defined(CLICK_OSKIT)

Packet::Packet()
{
  panic("Packet constructor");
}

Packet::~Packet()
{
  panic("Packet destructor");
}

Packet *
Packet::make(unsigned headroom, const unsigned char *data, unsigned len,
	     unsigned tailroom)
{
  unsigned size = len + headroom + tailroom;
  struct sk_buff *skb = alloc_skb(size, GFP_ATOMIC);
  if (skb) {
    skb_reserve(skb, headroom);	// leave some headroom
    skb_put(skb, len);		// leave space for data
    if (data) memcpy(skb->data, data, len);
  } else {
    click_chatter("oops, kernel could not allocate memory for skbuff");
    return 0;
  }
  Packet *p = (Packet *) skb;
  p->set_mac_broadcast_anno(0);
  p->set_fix_ip_src_anno(0);
  p->set_color_anno(0);
  p->set_siblings_anno(0);
#ifdef __KERNEL__
  p->set_cycle_anno(0,0);
  p->set_cycle_anno(1,0);
  p->set_cycle_anno(2,0);
  p->set_cycle_anno(3,0);
#endif
#ifdef MEASURE
  p->anno()->stats._measure = p->anno()->stats._created = 0;
#ifdef PERFMON
  p->anno()->stats._counter0 = p->anno()->stats._counter1 = 0;
#endif
#endif
  return(p);
}

#else /* __KERNEL__ */

inline
Packet::Packet()
{
  _use_count = 1;
  _data_packet = 0;
  _head = _data = _tail = _end = 0;
#ifdef CLICK_NODEOS
  _pbuf = 0;
  _chan = 0;
  _external = 0;
#endif
#ifdef MEASURE
  anno()->stats._measure = anno()->stats._created = 0;
#ifdef PERFMON
  anno()->stats._counter0 = anno()->stats._counter1 = 0;
#endif
#endif
  memset(_cb, 0, sizeof(_cb));
}

Packet::~Packet()
{
  if (_data_packet) {
    _data_packet->kill();
  } else if (_head) {
#ifdef CLICK_NODEOS
    if (_pbuf) {
      if (_external) {
        ani_pbuf_deinit(_pbuf);
	ani_mempool_free(ani_mempool_current(), _head, _end - _head);
      }
      else {
	if (_chan) {
	  assert(ani_pbuf_bufsize(_pbuf) == NODEOS_CHAN_BUFSIZE);
	  ani_chan_pbuf_add(_chan, &_pbuf, 1);
	} else
	  panic("Packet::~Packet");
      }
    }
    else
	ani_mempool_free(ani_mempool_current(), _head, _end - _head);
#else
    delete[] _head;
#endif
    _head = _data = 0;
  }
}

inline Packet *
Packet::make(int, int, int)
{
  return new Packet(6, 6, 6);
}

void
Packet::alloc_data(unsigned headroom, unsigned len, unsigned tailroom)
{
  unsigned n = len + headroom + tailroom;
#ifdef CLICK_NODEOS
  _head = (unsigned char *) ani_mempool_alloc(ani_mempool_current(), n);
  _external = 1;
  if (!_head)
    click_chatter("Packet::alloc_data no more memory");
#else
  _head = new unsigned char[n];
#endif
  _data = _head + headroom;
  _tail = _data + len;
  _end = _head + n;
}

Packet *
Packet::make(unsigned headroom, const unsigned char *data, unsigned len,
	     unsigned tailroom)
{
  Packet *p = new Packet;
  if (p) {
    p->alloc_data(headroom, len, tailroom);
    if (data && p->data()) memcpy(p->data(), data, len);
  }
  return p;
}

#ifdef CLICK_NODEOS
Packet *
Packet::make(ani_pbuf_t *pbuf, ani_chan_t *chan)
{
  Packet *p = new Packet;

  /*
   * Form a Click Packet from the internal data buffer. We are very lucky
   * that the nodeos pbuf format is so simple. If it becomes some kind of
   * scatter gather thing, we will need to get more clever (or less clever
   * and copy it).
   */
  p->_pbuf = pbuf;
  p->_chan = chan;
  p->_head  = (unsigned char *) ani_pbuf_bufdata(pbuf);
  p->_data  = (unsigned char *) ani_pbuf_data(pbuf);
  p->_tail  = p->_data + ani_pbuf_size(pbuf);
  p->_end   = p->_head + ani_pbuf_bufsize(pbuf);
  p->_external = 0;

  return p;
}
#endif

#endif /* __KERNEL__ */

//
// UNIQUEIFICATION
//

#if defined(__KERNEL__) || defined(CLICK_OSKIT)

Packet *
Packet::clone()
{
  struct sk_buff *n = skb_clone(skb(), GFP_ATOMIC);
  return (Packet *)n;
}

Packet *
Packet::uniqueify_copy()
{
  struct sk_buff *n = skb_copy(skb(), GFP_ATOMIC);
  // all annotations, including IP header annotation, are copied,
  // but IP header will point to garbage if old header was 0
  if (!ip_header()) n->nh.iph = 0;
  kill();
  return (Packet *)n;
}

#else /* user level */

Packet *
Packet::clone()
{
  Packet *p = Packet::make(6, 6, 6); // dummy arguments: no initialization
  if (!p) return 0;
  p->_use_count = 1;
  p->_data_packet = (Packet *)this;
  p->_head = _head;
  p->_data = _data;
  p->_tail = _tail;
  p->_end = _end;
  p->copy_annotations(this);
  p->_nh_iph = _nh_iph;
  p->_h_raw = _h_raw;
  // increment our reference count because of _data_packet reference
  _use_count++;
  return p;
}

Packet *
Packet::uniqueify_copy()
{
  Packet *p = Packet::make(6, 6, 6); // dummy arguments: no initialization
  if (!p) return 0;
  p->_use_count = 1;
  p->_data_packet = 0;
  p->alloc_data(headroom(), length(), tailroom());
  memcpy(p->_data, _data, _tail - _data);
  p->copy_annotations(this);
  if (_nh_iph) {
    p->_nh_iph = (click_ip *)(p->_data + ip_header_offset());
    p->_h_raw = (unsigned char *)(p->_data + transport_header_offset());
  } else {
    p->_nh_iph = 0;
    p->_h_raw = 0;
  }
  memcpy(p->_cb, _cb, sizeof(_cb));
  kill();
  return p;
}

#endif


//
// PUSH AND PULL
//

/*
 * Prepend some empty space before a packet.
 * May kill this packet and return a new one.
 */
Packet *
Packet::expensive_push(unsigned int nbytes)
{
  click_chatter("expensive Packet::push");
#ifdef __KERNEL__
  Packet *q = Packet::make(skb_realloc_headroom(skb(), nbytes));
  // oops! -- patch from Richard Mortier
  (void)skb_push(q->skb(), nbytes);
#elif defined(CLICK_OSKIT)
  Packet *q = 0;
  panic("Can't do it!\n");
#else
  Packet *q = Packet::make(length() + nbytes);
  memcpy(q->data() + nbytes, data(), length());
  memcpy(q->anno(), anno(), sizeof(Anno));
#endif
  kill();
  return q;
}

Packet *
Packet::put(unsigned int nbytes)
{
  if (tailroom() >= nbytes) {
    Packet *p = uniqueify();
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
    skb_put(p->skb(), nbytes);
#else
    p->_tail += nbytes;
#ifdef CLICK_NODEOS
    if (ani_pbuf_setsize(p->_pbuf, p->_tail - p->_data, 0) != 0)
      panic("Packet::put");
#endif
#endif
    return p;
  } else {
    click_chatter("expensive Packet::put");
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
    Packet *q = 0;
    panic("Packet::put");
#else
    Packet *q = Packet::make(length() + nbytes);
    memcpy(q->data(), data(), length());
    memcpy(q->anno(), anno(), sizeof(Anno));
#endif
    kill();
    return q;
  }
}

Packet *
Packet::take(unsigned int nbytes)
{
  if(nbytes <= length()){
    Packet *p = uniqueify();
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
    skb()->tail -= nbytes;
    skb()->len -= nbytes;
#else
    p->_tail -= nbytes;
#ifdef CLICK_NODEOS
    if (ani_pbuf_setsize(p->_pbuf, p->_tail - p->_data, 0) != 0)
      panic("Packet::take");
#endif
#endif    
    return(p);
  } else {
    click_chatter("Packet::take oops");
    return(this);
  }
}
