#ifndef PACKET_HH
#define PACKET_HH
#include "ipaddress.hh"
#include "glue.hh"
struct click_ip;

#ifdef CLICK_OSKIT
extern "C" {
#include <oskit/io/skbufio.h>
}
#endif
#ifdef CLICK_NODEOS
extern "C" {
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_pbuf.h>
#include <nodeos/impl/ani_mempool.h>
}
#endif

class Packet {

  // Anno must fit in sk_buff's char cb[48].
  struct Anno {
    IPAddress dst_ip;
    unsigned char ip_tos;
    unsigned char ip_ttl;
    unsigned short ip_off;
    char mac_broadcast; // flag: MAC address was a broadcast or multicast.
    char fix_ip_src;    // flag: asks FixIPSrc to set ip_src.
    char param_off;     // for ICMP Parameter Problem, byte offset of error.
    char color;         // one of 255 colors set by Paint element.
    int siblings;       // written by IPFlexMonitor, read by Block
#ifdef __KERNEL__
    union {
      cycles_t cycles[4];
      struct {
	unsigned m0[2];
	unsigned m1[2];
      } perf;
    } p;
#endif
#ifdef MEASURE
    struct {
	    char		_measure;
	    unsigned long long	_created;
#ifdef PERFMON
	    unsigned long long	_counter0;
	    unsigned long long	_counter1;
#endif
    } stats;
#endif
  };
  
#if !defined(__KERNEL__) && !defined(CLICK_OSKIT)
  int _use_count;
  Packet *_data_packet;
  /* mimic Linux sk_buff */
  unsigned char *_head; /* start of allocated buffer */
  unsigned char *_data; /* where the packet starts */
  unsigned char *_tail; /* one beyond end of packet */
  unsigned char *_end;  /* one beyond end of allocated buffer */
  unsigned char _cb[48];
  click_ip *_nh_iph;
  unsigned char *_h_raw;
#ifdef CLICK_NODEOS
  ani_pbuf_t	*_pbuf;
  ani_chan_t	*_chan;
  bool		_external;
#endif
#endif
  
  Packet();
  Packet(const Packet &);
  ~Packet();
  Packet &operator=(const Packet &);

#if !defined(__KERNEL__) && !defined(CLICK_OSKIT)
  Packet(int, int, int)			{ }
  static Packet *make(int, int, int);
  void alloc_data(unsigned, unsigned, unsigned);
#endif

  Packet *uniqueify_copy();

  Packet *expensive_push(unsigned int nbytes);
  
#if !defined(__KERNEL__) && !defined(CLICK_OSKIT)
  const Anno *anno() const		{ return (const Anno *)_cb; }
  Anno *anno()				{ return (Anno *)_cb; }
#endif
  
  friend class ShutUpCompiler;
  
 public:
  static unsigned default_headroom()	{ return 28; }
  static unsigned default_tailroom(unsigned len) { return (len<56?64-len:8); }
  
  static Packet *make(unsigned);
  static Packet *make(const char *, unsigned);
  static Packet *make(const unsigned char *, unsigned);
  static Packet *make(unsigned, const unsigned char *, unsigned, unsigned);
  
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  /*
   * Wraps a Packet around an existing sk_buff.
   * Packet now owns the sk_buff (ie we don't increment skb->users).
   */
  static Packet *make(struct sk_buff *);
  struct sk_buff *skb() const		{ return (struct sk_buff *)this; }
  struct sk_buff *steal_skb()		{ return skb(); }

 private:
  const Anno *anno() const		{ return (const Anno *)skb()->cb; }
  Anno *anno()				{ return (Anno *)skb()->cb; }
#endif

 public:
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  void kill() 				{ __kfree_skb(skb()); }
#else
  void kill()				{ if (--_use_count <= 0) delete this; }
#endif

#ifdef CLICK_NODEOS
  static Packet *make(ani_pbuf_t *p, ani_chan_t *c);
  ani_pbuf_t *nodeos_packet(void) const { return _pbuf; }
  void set_nodeos_packet(ani_pbuf_t *p) { assert(_external == 1); _pbuf = p; }
#endif

  bool shared() const;
  Packet *clone();
  Packet *uniqueify();
  
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  unsigned char *data() const	{ return skb()->data; }
  unsigned length() const	{ return skb()->len; }
  unsigned headroom() const	{ return skb()->data - skb()->head; }
  unsigned tailroom() const	{ return skb()->end - skb()->tail; }
#else
  unsigned char *data() const	{ return _data; }
  unsigned length() const	{ return _tail - _data; }
  unsigned headroom() const	{ return _data - _head; }
  unsigned tailroom() const	{ return _end - _tail; }
#endif
  
  Packet *push(unsigned nbytes);	// Add more space before packet.
  Packet *nonunique_push(unsigned);
  void pull(unsigned nbytes);		// Get rid of initial bytes.
  Packet *put(unsigned nbytes);		// Add bytes to end of pkt.
  Packet *take(unsigned nbytes);	// Delete bytes from end of pkt.

#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  click_ip *ip_header() const		{ return (click_ip *)skb()->nh.iph; }
  unsigned char *transport_header() const	{ return skb()->h.raw; }
#else
  click_ip *ip_header() const		{ return _nh_iph; }
  unsigned char *transport_header() const	{ return _h_raw; }
#endif
  void set_ip_header(click_ip *, unsigned);
  unsigned ip_header_offset() const;
  unsigned ip_header_length() const;
  unsigned transport_header_offset() const;

  void copy_annotations(Packet *);
  
  void set_dst_ip_anno(IPAddress a)	{ anno()->dst_ip = a; }
  IPAddress dst_ip_anno() const		{ return anno()->dst_ip; }
  
  void set_ip_tos_anno(unsigned char t)	{ anno()->ip_tos = t; }
  unsigned char ip_tos_anno() const	{ return anno()->ip_tos; }
  void set_ip_ttl_anno(unsigned char t)	{ anno()->ip_ttl = t; }
  unsigned char ip_ttl_anno() const	{ return anno()->ip_ttl; }
  void set_ip_off_anno(unsigned short o){ anno()->ip_off = o; }
  unsigned short ip_off_anno() const    { return anno()->ip_off; }
  void set_mac_broadcast_anno(char b)	{ anno()->mac_broadcast = b; }
  char mac_broadcast_anno() const	{ return anno()->mac_broadcast; }
  void set_fix_ip_src_anno(char f)	{ anno()->fix_ip_src = f; }
  char fix_ip_src_anno() const		{ return anno()->fix_ip_src; }
  void set_param_off_anno(char p)	{ anno()->param_off = p; }
  char param_off_anno() const		{ return anno()->param_off; }
  void set_color_anno(char c)		{ anno()->color = c; }
  char color_anno() const		{ return anno()->color; }
  void set_siblings_anno(int s)	        { anno()->siblings = s; }
  int siblings_anno() const		{ return anno()->siblings; }
#ifdef __KERNEL__
  void set_cycle_anno(int i, cycles_t v) { anno()->p.cycles[i] = v; }
  void set_metric0_anno(int i, unsigned v) { anno()->p.perf.m0[i] = v; }
  void set_metric1_anno(int i, unsigned v) { anno()->p.perf.m1[i] = v; }
  cycles_t cycle_anno(int i) const	{ return anno()->p.cycles[i]; }
  unsigned metric0_anno(int i) const	{ return anno()->p.perf.m0[i]; }
  unsigned metric1_anno(int i) const	{ return anno()->p.perf.m1[i]; }
#endif
#ifdef MEASURE
  void set_measure_measure(void)	{ anno()->stats._measure = 1; }
  int  get_measure_measure(void) const	{ return anno()->stats._measure; }
  void set_measure_created(unsigned long long created)
					{ anno()->stats._created = created; }
  unsigned long long get_measure_created(void) const
					{ return anno()->stats._created; }
#ifdef PERFMON
  void set_measure_counter0(unsigned long long counter0)
					{ anno()->stats._counter0 = counter0; }
  unsigned long long get_measure_counter0(void) const
					{ return anno()->stats._counter0; }
  void set_measure_counter1(unsigned long long counter1)
					{ anno()->stats._counter1 = counter1; }
  unsigned long long get_measure_counter1(void) const
					{ return anno()->stats._counter1; }
#endif
#endif
};


inline Packet *
Packet::make(unsigned len)
{
  return make(default_headroom(), (const unsigned char *)0, len,
	      default_tailroom(len));
}

inline Packet *
Packet::make(const char *s, unsigned len)
{
  return make(default_headroom(), (const unsigned char *)s, len,
	      default_tailroom(len));
}

inline Packet *
Packet::make(const unsigned char *s, unsigned len)
{
  return make(default_headroom(), (const unsigned char *)s, len,
	      default_tailroom(len));
}

#ifdef __KERNEL__
inline Packet *
Packet::make(struct sk_buff *skb)
{
  if (atomic_read(&skb->users) == 1)
    return (Packet *)skb;
  else
    return (Packet *)skb_clone(skb, GFP_ATOMIC);
}
#endif

#ifdef CLICK_OSKIT
inline Packet *
Packet::make(struct sk_buff *skb)
{
  /*
   * I am going to stipulate that an incoming packet from the oskit
   * does not need to be cloned.
   */
  memset(skb->cb, 0, sizeof(skb->cb));
  return (Packet *)skb;
}
#endif

inline bool
Packet::shared() const
{
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  return skb_cloned(skb());
#else
  return (_data_packet || _use_count > 1);
#endif
}

inline Packet *
Packet::uniqueify()
{
  if (shared())
    return uniqueify_copy();
  else
    return this;
}

inline Packet *
Packet::push(unsigned int nbytes)
{
  if (headroom() >= nbytes) {
    Packet *p = uniqueify();
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
    skb_push(p->skb(), nbytes);
#else
    p->_data -= nbytes;
#ifdef CLICK_NODEOS
    if (ani_pbuf_setoffset(p->_pbuf, p->_data - p->_head, 0) != 0)
      panic("Packet::push");
#endif
#endif
    return p;
  } else
    return expensive_push(nbytes);
}

inline Packet *
Packet::nonunique_push(unsigned int nbytes)
{
  if (headroom() >= nbytes) {
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
    skb_push(skb(), nbytes);
#else
    _data -= nbytes;
#endif
    return this;
  } else
    return expensive_push(nbytes);
}

/*
 * Get rid of some bytes at the start of a packet.
 */
inline void
Packet::pull(unsigned int nbytes)
{
  if (nbytes > length()) {
    click_chatter("Packet::pull %d > length %d\n", nbytes, length());
    nbytes = length();
  }
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  skb_pull(skb(), nbytes);
#else
  _data += nbytes;
#ifdef CLICK_NODEOS
  if (ani_pbuf_setoffset(_pbuf, _data - _head, 0) != 0)
    panic("Packet::pull");
#endif
#endif
}

inline void
Packet::set_ip_header(click_ip *iph, unsigned len)
{
#if defined(__KERNEL__) || defined(CLICK_OSKIT)
  skb()->nh.iph = (struct iphdr *)iph;
  skb()->h.raw = (unsigned char *)iph + len;
#else
  _nh_iph = iph;
  _h_raw = (unsigned char *)iph + len;
#endif
}

inline unsigned
Packet::ip_header_offset() const
{
  return (unsigned char *)ip_header() - data();
}

inline unsigned
Packet::ip_header_length() const
{
  return (unsigned char *)transport_header() - (unsigned char *)ip_header();
}

inline unsigned
Packet::transport_header_offset() const
{
  return (unsigned char *)transport_header() - data();
}

inline void
Packet::copy_annotations(Packet *p)
{
  *anno() = *p->anno();
}

#endif
