/*
 * timer.{cc,hh} -- portable timers. Linux kernel module uses Linux timers
 * Eddie Kohler
 *
 * Copyright (c) 1999-2000 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYRIGHT' file in the source
 * distribution.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "timer.hh"
#include "element.hh"
#include "router.hh"
#ifdef CLICK_NODEOS
extern "C" {
#include <nodeos/impl/ani_event.h>
}
#endif

Timer::Timer()
  : _prev(this), _next(this), _hook(head_hook), _thunk(0), _head(this)
{
}

Timer::Timer(Timer *head, TimerHook hook, unsigned long thunk)
  : _prev(0), _next(0), _hook(hook), _thunk(thunk), _head(head)
{
  assert(_head->is_head());
}

Timer::Timer(TimerHook hook, unsigned long thunk)
  : _prev(0), _next(0), _hook(hook), _thunk(thunk), _head(0)
{
}

Timer::Timer(Element *e)
  : _prev(0), _next(0), _hook(element_hook), _thunk((unsigned long)e), _head(0)
{
}

void
Timer::head_hook(unsigned long)
{
  assert(0);
}

/*
 * Timer::element_hook is a callback that gets called when a Timer,
 * constructed with just an Element instance, expires. 
 * 
 * When used in userlevel or kernel polling mode, timer is maintained by
 * Click, so Timer::element_hook is called within Click.
 */

void
Timer::element_hook(unsigned long thunk)
{
  Element *e = (Element *)thunk;
  /* put itself on the work list */
  e->schedule_immediately();	// might not have tickets
}

void
Timer::attach(Router *r)
{
  assert(!attached());
#ifdef CLICK_NODEOS
  _router = r;
#endif
  attach(r->timer_head());
}

void
Timer::attach(Element *e)
{
  assert(!attached());
#ifdef CLICK_NODEOS
  _router = e->router();
#endif
  attach(e->router()->timer_head());
}

void
Timer::schedule_after_ms(int ms)
{
  assert(!is_head() && attached());
  if (scheduled())
    unschedule();
  click_gettimeofday(&_expires);
  struct timeval interval;
  interval.tv_sec = ms / 1000;
  interval.tv_usec = (ms % 1000) * 1000;
  timeradd(&_expires, &interval, &_expires);
  Timer *prev = _head;
  Timer *trav = prev->_next;
  while (trav != _head && timercmp(&_expires, &trav->_expires, >)) {
    prev = trav;
    trav = trav->_next;
  }
  _prev = prev;
  _next = trav;
  _prev->_next = this;
  trav->_prev = this;
#ifdef CLICK_NODEOS
  assert(_router);
  _router->nodeos_schedule_event(&_expires);
#endif
}

void
Timer::unschedule()
{
  if (scheduled()) {
    _prev->_next = _next;
    _next->_prev = _prev;
    _prev = _next = 0;
  }
}

void
Timer::run()
{
  assert(is_head());
  struct timeval now;
  click_gettimeofday(&now);
  while (_next != this && !timercmp(&_next->_expires, &now, >)) {
    Timer *t = _next;
    _next = t->_next;
    _next->_prev = this;
    t->_prev = 0;
    t->_hook(t->_thunk);
  }
#ifdef CLICK_NODEOS
  if (_next != this) {
    assert(_router != 0);
    _router->nodeos_schedule_event(&_next->_expires);
  } else if (_router)
    _router->nodeos_unschedule_event();
#endif
}

// How long select() should wait until next timer expires.
int
Timer::get_next_delay(struct timeval *tv)
{
  assert(is_head());
  if (_next == this) {
    tv->tv_sec = 1000;
    tv->tv_usec = 0;
    return 0;
  } else {
    struct timeval now;
    click_gettimeofday(&now);
    if(timercmp(&_next->_expires, &now, >)){
      timersub(&_next->_expires, &now, tv);
    } else {
      tv->tv_sec = 0;
      tv->tv_usec = 0;
    }
    return 1;
  }  
}

