/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "lexer.hh"
#include "router.hh"
#include "error.hh"
#include "timer.hh"
#include "straccum.hh"
#include "elements/nodeos/frominchan.hh"
#include "elements/nodeos/tooutchan.hh"

extern "C" {
#include <nodeos/an_error.h>
#include <nodeos/an_resource.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_demuxkey.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_event.h>
}

/*
 * This is in the auto generated elements.cc file.
 */
extern void export_elements(Lexer *);

extern "C" an_chan_t *
an_clickchan_create(void *mem, an_flow_t *flow, char *config,
		    an_chan_t *inchan, an_chan_t *outchan)
{
	ani_error_t	err;
	ani_flow_t	*iflow = 0;
	ani_chan_t	*ichan, *iinchan, *ioutchan;
	ani_mempool_t	*cpool = 0;
	int		didalloc = 0;
	an_cpuspec_t	cspec;
	static int	initialized;

	if (! initialized) {
		/*
		 * XXX Anything static is bound to be some kind of
		 * problem with multiple router graphs in operation.
		 */
		String::static_initialize();
		ErrorHandler::static_initialize(
			new FileErrorHandler(stderr, ""));

		initialized = 1;
	}

	/*
	 * Leave these here. Its better if the C++ initializers above
	 * are called first, before declaring C++ variables.
	 */
	Element	        *elem;
	Lexer		*lex    = 0;
	Router		*router = 0;
	ErrorHandler	*errh   = ErrorHandler::default_handler();
	String		graphstr;
	String		filename;
	int		cookie;

	/*
	 * First off, we need to create the graph.
	 */
#ifdef  CLICK_GRAPH_FILE
	FILE		*fp;
	StringAccum	config_sa;

	if ((fp = fopen(config, "r")) == NULL) {
		errh->message("Could not open config file: %s:", config);
		return 0;
	}

	/*
	 * So, not sure why this is like this. Must read the file first and
	 * put it into String format.
	 */
	while (!feof(fp)) {
		char	*x;
		int	r;

		if ((x = config_sa.reserve(512)) == NULL)
			errh->fatal("Out of memory while reading config file");
		else {
			r = fread(x, 1, 512, fp);
			config_sa.forward(r);
		}
	}
	fclose(fp);
	graphstr = config_sa.take_string;
	filename = config;
#else
	graphstr = config;
	filename = "inline";
#endif
	/*
	 * Do all the lexer goo (parse config file).
	 */
        // Prepare the lexer
	lex = new Lexer(errh);
	export_elements(lex);

	// Now run the lexer.
	cookie = lex->begin_parse(graphstr, filename, 0);
	while (lex->ystatement())
		/* do nothing */;
  
	// Create the router from the lexed configuration.
	router = lex->create_router();
	lex->end_parse(cookie);

	/*
	 * Initialize the router graph.
	 */
	if (router->initialize(errh) < 0) {
		errh->message("Could not initialize the router");
		delete router;
		delete lex;
		return 0;
	}

#if 1
	/*
	 * Show what we did.
	 */
	errh->message("Router configuration structure:");
	router->print_structure(errh);
#endif

	/*
	 * Okay, we have a router graph. Since that all worked okay, lets
	 * do the nodeos specific goo.
	 */
	NODEOS_ENTER;

	err = ani_error_clear();

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(ani_flow_current());
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_chan_t));
		didalloc = 1;
	}

	/*
	 * Validate arguments.
	 */
	ichan = (ani_chan_t *) NODEOS_INTERN_CHAN((an_chan_t *)mem);
	iflow = NODEOS_INTERN_FLOW(flow);
	iinchan = NODEOS_INTERN_CHAN(inchan);
	ioutchan = NODEOS_INTERN_CHAN(outchan);
	if (iflow == 0 || iinchan == 0 || ioutchan == 0 ||
	    ani_chan_type(iinchan) != CHAN_IN || iinchan->tdata.in.func != 0 ||
	    ani_chan_type(ioutchan) != CHAN_OUT) {
		errh->message("Could not allocate channel threadpool");
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid parameter");
		goto done;
	}

	/*
	 * XXX create a threadpool with a single thread to serialize packets
	 * and events for the Click graph.
	 */
	cspec.ttype = AN_TH_IMPLICIT;
	cspec.threads.im.maxthreads = 1;
	cspec.threads.im.stacksize = 0;
	err = ani_threadpool_init(router->nodeos_tpool(), iflow, &cspec);
	if (ani_error_occured(err)) {
		errh->message("Could not create channel threadpool");
		goto done;
	}

	/*
	 * We need to hook up the endpoints of the graph (FromInChan
	 * and ToInchan) to the channel objects.
	 */
	elem = 0;
	for (int i = 0; i < router->nelements(); i++) {
		if (! strcmp(router->element(i)->class_name(), "ToOutChan")) {
			elem = router->element(i);
			break;
		}
	}
	if (elem == 0) {
		errh->message("Could not find the ToOutChan element");
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid outchan element");
		goto done;
	}
	((ToOutChan *)elem)->set_chan(ioutchan);

	elem = 0;
	for (int i = 0; i < router->nelements(); i++) {
		if (! strcmp(router->element(i)->class_name(), "FromInChan")) {
			elem = router->element(i);
			break;
		}
	}
	if (elem == 0) {
		errh->message("Could not find the FromInChan element");
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid inchan element");
		goto done;
	}
	((FromInChan *)elem)->set_chan(iinchan);
	
	/*
	 * Create and enable the Nodeos cutchannel. 
	 */
	err = ani_cutchan_init(ichan, iflow, router->nodeos_tpool(), router, 0,
			       iinchan, ioutchan, click_FromInChan_in, elem);
	if (!ani_error_occured(err)) {
		err = ani_chan_enable(iinchan);
		if (ani_error_occured(err))
			ani_chan_deinit(ichan);
	}
	if (ani_error_occured(err)) {
		errh->message("Could not create cut channel");
		goto done;
	}

	/*
	 * I guess we are done.
	 */
 done:
	if (ani_error_occured(err)) {
		if (didalloc)
			ani_mempool_free(cpool, mem, sizeof(an_chan_t));
		if (router) {
			delete router;
			delete lex;
		}
			
	}

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_CHAN(ichan);
}

extern "C" an_error_t
an_clickchan_destroy(an_chan_t *chan)
{
	struct ani_chan *ichan;
	struct ani_threadpool *tpool;
	Router *router;
	ani_error_t err;

	NODEOS_ENTER;

	ichan = NODEOS_INTERN_CHAN(chan);
	if (ichan == 0) {
		NODEOS_LEAVE;
		return AN_ERR_INVALIDARG;
	}

	router = (Router *)ichan->data;
	tpool = router->nodeos_tpool();

	err = ani_chan_deinit(ichan);
	if (!ani_error_occured(err)) {
		assert_ani_threadpool_valid(tpool);
		ani_threadpool_disable(tpool, 1);
		ani_threadpool_deinit(tpool, 1);

		delete router;
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}
