/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Click COM component.
 */

extern "C" {
#include <oskit/c/malloc.h>
#include "click_com.h"
}

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "lexer.hh"
#include "router.hh"
#include "error.hh"
#include "timer.hh"
#include "straccum.hh"
#include "archive.hh"
#include <unistd.h>
#include <dlfcn.h>

/*
 * XXX Belongs elsewhere!
 */
const struct oskit_guid oskit_clicker_iid = OSKIT_CLICKER_IID;

/*
 * This is in the auto generated elements.cc file.
 */
extern void export_elements(Lexer *);

/*
 * Current lexer for dyn loads.
 */
static Lexer	*CurLexer;
static int	load_package(String package, ErrorHandler *errh);
extern "C" {
extern void	oskit_boot_rtld(char *);
}

#ifdef MEASURE
extern int			packet_count;
extern int			packet_total;
#if    defined(CLICK_OSKIT) && defined(PERFMON)
#include <oskit/machine/perfmon.h>
extern unsigned long long	perfmon_counter0;
extern unsigned long long	perfmon_counter1;
#endif
#endif /* MEASURE */

#ifdef MULTITHREADED
/*
 * Need a lock to protect initialization and finalization of routers.
 */
static pthread_mutex_t		mymutex = PTHREAD_MUTEX_INITIALIZER;
#define LOCKIT()		pthread_mutex_lock(&mymutex)
#define UNLOCKIT()		pthread_mutex_unlock(&mymutex)
#else
#define LOCKIT()
#define UNLOCKIT()
#endif

/*
 * The actual component. 
 */
struct clicker {
	oskit_clicker_t		ioi;		/* The COM interface */
	int			count;		/* Ref count */
	Lexer		       *lex;
	Router		       *router;
	int			stopped;
#ifdef  MULTITHREADED
	pthread_t		ptid;
	pthread_attr_t		thread_attrs;
	bool			valid_attrs;
#endif
};

static OSKIT_COMDECL
clickcon_query(oskit_clicker_t *c,
	       const oskit_iid_t *iid, void **out_ihandle) 
{
	struct clicker	 *clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_clicker_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &clicker->ioi;
		++clicker->count;
		
		return 0;
	}

	*out_ihandle = NULL;
	return OSKIT_E_NOINTERFACE;
}

static OSKIT_COMDECL_U
clickcon_addref(oskit_clicker_t *c)
{
	struct clicker	 *clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	return ++clicker->count;
}

static OSKIT_COMDECL_U
clickcon_release(oskit_clicker_t *c)
{
	struct clicker	 *clicker = (struct clicker *) c;
	unsigned newcount;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if ((newcount = --clicker->count) == 0) {
#ifdef  MULTITHREADED
		void	*stat;
		
		/*
		 * Cancel it before deleting it!
		 */
		pthread_cancel(clicker->ptid);
		pthread_join(clicker->ptid, &stat);
		LOCKIT();
#endif
		delete clicker->router;
		delete clicker->lex;
		sfree(clicker, sizeof(*clicker));
		UNLOCKIT();
	}
	return newcount;
}

static OSKIT_COMDECL
clickcon_config(oskit_clicker_t *c, char *config_file)
{
	struct clicker	*clicker = (struct clicker *) c;
	ErrorHandler	*errh    = ErrorHandler::default_handler();
	FILE		*fp;
	String		filename = config_file;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if ((fp = fopen(config_file, "r")) == NULL)
		errh->fatal("Could not open config file: %s", config_file);

	/*
	 * So, not sure why this like this. Must read the file first and
	 * put it into String format.
	 */
	StringAccum	config_sa;
	while (!feof(fp)) {
		char	*x;
		int	r;

		if ((x = config_sa.reserve(512)) == NULL)
			errh->fatal("Out of memory");
		else {
			r = fread(x, 1, 512, fp);
			config_sa.forward(r);
		}
	}
	fclose(fp);

        // Prepare the lexer
	Lexer *lex = new Lexer(errh);
	export_elements(lex);

	/*
	 * Dyn load support. This is taken from userlevel/click.cc
	 */
	CurLexer = lex;

	String config_str = config_sa.take_string();
	if (config_str.length() != 0 && config_str[0] == '!') {
		Vector<ArchiveElement> archive;
		separate_ar_string(config_str, archive, errh);
		bool found_config = false;
		for (int i = 0; i < archive.size(); i++)
			if (archive[i].name == "config") {
				config_str = archive[i].data;
				found_config = true;
			}
			else if (archive[i].name.length() > 3 &&
				 archive[i].name.substring(-3) == ".uo") {

				String tmpnam = "foo.uo";
				
				FILE *f = fopen(tmpnam.cc(), "wb");
				fwrite(archive[i].data.data(), 1,
				       archive[i].data.length(), f);
				fclose(f);
				int ok = load_package(tmpnam, errh);
				unlink(tmpnam.cc());
				if (ok < 0)
					exit(1);
			}
		if (!found_config) {
			errh->error("archive has no `config' section");
			exit(1);
		}
	}

	// Now run the lexer.
	int cookie = lex->begin_parse(config_str, filename, 0);
	while (lex->ystatement())
		/* do nothing */;

	// Create the router from the lexed configuration.
	Router *router = lex->create_router();
	lex->end_parse(cookie);

	clicker->lex    = lex;
	clicker->router = router;

	return 0;
}

/*
 * Click driver runs in its own thread
 */
static void *
clickcon_rundriver(void *arg)
{
	struct clicker	*clicker = (struct clicker *) arg;
	Router  	*router  = clicker->router;
	ErrorHandler	*errh    = ErrorHandler::default_handler();

	LOCKIT();
	if (router->initialize(errh) >= 0) {
		errh->message("Router configuration structure:");
		router->print_structure(errh);
		UNLOCKIT();

		router->driver();
	}
	else
		UNLOCKIT();
	
	clicker->stopped = 1;
		
#ifdef  MEASURE
	if (packet_count == 0)
		goto skipit;
	
	printf("Time: %d packets at %d cycles per packet\n",
	       packet_count, packet_total / packet_count);
#if     defined(CLICK_OSKIT) && defined(PERFMON)
	printf("Perfmon: %s %d (%d per packet)\n", COUNTER0_STRING,
	       (int) perfmon_counter0,
	       (int) (perfmon_counter0 / packet_count));
	printf("Perfmon: %s %d (%d per packet)\n", COUNTER1_STRING,
	       (int) perfmon_counter1,
	       (int) (perfmon_counter1 / packet_count));
#endif
 skipit:
#endif
	return 0;
}

static OSKIT_COMDECL
clickcon_start(oskit_clicker_t *c)
{
	struct clicker	*clicker = (struct clicker *) c;
	Router  	*router;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	router = clicker->router;

#if     defined(MEASURE) && (defined(CLICK_OSKIT) && defined(PERFMON))
#if 1
	enable_perfcounter(0, PMC6_INST_RETIRED,         0, PMCF_OS|PMCF_USR);
#define COUNTER0_STRING "Inst Retired: " 
	enable_perfcounter(1, PMC6_UOPS_RETIRED,         0, PMCF_OS|PMCF_USR);
#define COUNTER1_STRING "UOPs Retired: "
#endif
#if 0
	enable_perfcounter(0, PMC6_BTB_MISSES,           0, PMCF_OS|PMCF_USR);
#define COUNTER0_STRING "BTB Misses:  " 
	enable_perfcounter(1, PMC6_DCU_MISS_OUTSTANDING, 0, PMCF_OS|PMCF_USR);
#define COUNTER1_STRING "DCU Miss Outstanding: "
#endif
#if 0
	enable_perfcounter(0, PMC6_BR_INST_RETIRED,      0, PMCF_OS|PMCF_USR);
#define COUNTER0_STRING "Branches Retired:  " 
	enable_perfcounter(1, PMC6_BR_MISS_PRED_RETIRED, 0, PMCF_OS|PMCF_USR);
#define COUNTER1_STRING "Branches MissPred: "
#endif
#if 0
	enable_perfcounter(0, PMC6_IFU_MEM_STALL,   0, PMCF_OS|PMCF_USR);
#define COUNTER0_STRING "IFU Mem Stall:" 
	enable_perfcounter(1, PMC6_IFU_IFETCH,      0, PMCF_OS|PMCF_USR);
#define COUNTER1_STRING "IFU Fetches:  "
#endif
#if 0
	enable_perfcounter(0, PMC6_IFU_IFETCH_MISS, 0, PMCF_OS|PMCF_USR);
#define COUNTER0_STRING "IFetch Misses:" 
	enable_perfcounter(1, PMC6_L2_IFETCH,     0xF, PMCF_OS|PMCF_USR);
#define COUNTER1_STRING "L2 Ifetches:  "
#endif
#endif
	
#ifdef  MULTITHREADED
	/*
	 * Click driver runs in its own thread. 
	 */
	int err = pthread_create(&clicker->ptid,
				 clicker->valid_attrs ?
				  &clicker->thread_attrs : 0,
				 clickcon_rundriver, (void *) c);
	if (err)
		return err;
#else
	clickcon_rundriver((void *) clicker);
#endif
	
	return 0;
}

static OSKIT_COMDECL_V
clickcon_stop(oskit_clicker_t *c)
{
	struct clicker	*clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	clicker->router->please_stop_driver();
}

static OSKIT_COMDECL_U
clickcon_stopped(oskit_clicker_t *c)
{
	struct clicker	*clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	return clicker->stopped;
}

static struct oskit_clicker_ops clickcon_ops = {
	clickcon_query,
	clickcon_addref,
	clickcon_release,
	clickcon_config,
	clickcon_start,
	clickcon_stop,
	clickcon_stopped,
};

/*
 * Create an instance of a click object. The object is pretty much useless
 * until it is configured, and then started.
 */
oskit_clicker_t *
oskit_clicker_create(void)
{
	struct clicker	*c;
	static int	initialized;

	if (! initialized) {
		String::static_initialize();
		ErrorHandler *errh = new FileErrorHandler(stderr, "");
		ErrorHandler::static_initialize(errh);

		/*
		 * Must initialize the RTLD library. 
		 */
		oskit_boot_rtld("kernel");

		initialized = 1;
	}

	c = (struct clicker *) smalloc(sizeof *c);
	if (c == 0)
		return 0;

	c->ioi.ops = &clickcon_ops;
	c->count   = 1;

#ifdef  MULTITHREADED
	c->stopped     = 0;
	c->valid_attrs = 0;
#endif
	return &c->ioi;
}

#ifdef MULTITHREADED
oskit_clicker_t *
oskit_clicker_create_threaded(pthread_attr_t *pattr)
{
	oskit_clicker_t *clicker;
	struct clicker	*c;

	if ((clicker = oskit_clicker_create()) == NULL)
		return NULL;

	c = (struct clicker *) clicker;

	/*
	 * Strictly speaking, copying the attributes structure is
	 * illegal since its definition is not only opaque but
	 * possibly meaningless to copy. But, since the spec provides
	 * no copy function, I'm gonna do it this way cause I know
	 * it will work in the oskit.
	 */
	memcpy(&c->thread_attrs, pattr, sizeof(*pattr));
	c->valid_attrs = 1;

	return clicker;
}
#endif

/*
 * This is for dynamic linking support. 
 */

/*
 * These first two are stubbed; they get called from the loaded file's
 * init code. We don't care about it.
 */
extern "C" void
click_provide(const char *)
{
}

extern "C" void
click_unprovide(const char *)
{
}

extern "C" int
click_add_element_type(const char *ename, Element *e)
{
	return CurLexer->add_element_type(ename, e);
}

extern "C" void
click_remove_element_type(int which)
{
	CurLexer->remove_element_type(which);
}

extern "C" {
	typedef int (*init_module_func)(void);
}

/*
 * This comes from userlevel/click.cc
 */
static int
load_package(String package, ErrorHandler *errh)
{
#ifndef RTLD_NOW
	void *handle = dlopen((char *)package.cc(), RTLD_LAZY);
#else
	void *handle = dlopen((char *)package.cc(), RTLD_NOW);
#endif
	if (!handle)
		return errh->error("cannot load package: %s", dlerror());
	
	void *init_sym = dlsym(handle, "init_module");
	if (!init_sym)
		return errh->error("package `%s' has no `init_module'",
				   package.cc());
	
	init_module_func init_func = (init_module_func)init_sym;
	if (init_func() != 0)
		return errh->error("error initializing package `%s'",
				   package.cc());
	
	return 0;
}
